package io.scanbot.sdk.flutter.camera

import android.content.Context
import android.graphics.Color
import android.graphics.PointF
import android.view.View
import android.widget.FrameLayout
import com.google.gson.FieldNamingPolicy
import com.google.gson.GsonBuilder
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.platform.PlatformView
import io.scanbot.sdk.ScanbotSDK
import io.scanbot.sdk.flutter.*
import io.scanbot.sdk.flutter.camera.config.CroppingResult
import io.scanbot.sdk.flutter.utils.convertDpToPx
import io.scanbot.sdk.persistence.Page
import io.scanbot.sdk.persistence.PageFileStorage
import io.scanbot.sdk.ui.EditPolygonImageView
import io.scanbot.sdk.ui.MagnifierView
import io.scanbot.sdk.ui.view.interactor.DetectPolygonOnPageUseCase
import io.scanbot.sdk.util.PolygonHelper
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.catch
import kotlinx.coroutines.flow.launchIn
import kotlinx.coroutines.flow.onEach
import kotlin.coroutines.CoroutineContext


class ScanbotCroppingPlatformView internal constructor(
    private val context: Context,
    private val messenger: BinaryMessenger,
    private val id: Int,
    private val creationParams: Json?
) : PlatformView,
    MethodChannel.MethodCallHandler, CoroutineScope {

    private val rootView: FrameLayout = FrameLayout(context)
    private val editPolygonImageView = EditPolygonImageView(context, null)
    private val magnifierView = MagnifierView(context, null)
    private val scanbotSDK = ScanbotSDK(context)
    private val contourDetector = scanbotSDK.createContourDetector()
    private val pageFileStorage = scanbotSDK.createPageFileStorage()
    private var croppingViewChannel = MethodChannel(messenger, "scanbot_sdk_cropping_$id")
    private var page: Page? = null
    private val returnContext = Dispatchers.Main

    val gson = GsonBuilder()
        .setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
        .registerTypeAdapter(ByteArray::class.java, ByteArrayToBase64TypeAdapter())
        .create()

    private var job: Job = Job()
    override val coroutineContext: CoroutineContext
        get() = Dispatchers.Default + job

    override fun getView(): View {
        return rootView
    }

    init {
        initView()
    }

    private fun initView() {
        rootView.addView(
            editPolygonImageView,
            FrameLayout.LayoutParams(
                FrameLayout.LayoutParams.MATCH_PARENT,
                FrameLayout.LayoutParams.MATCH_PARENT
            )
        )
        rootView.addView(
            magnifierView,
            FrameLayout.LayoutParams(
                FrameLayout.LayoutParams.MATCH_PARENT,
                FrameLayout.LayoutParams.MATCH_PARENT
            )
        )

        if (!scanbotSDK.licenseInfo.isValid) {
            launch(returnContext) {
                val json = scanbotSDK.getLicenseStatusJson()
                croppingViewChannel.invokeMethod(
                    "onLicenseError",
                    json.toString()
                )
            }
        }

        creationParams?.checkParam<Json>("page") {
            val page = it.toPage(gson)
            this.page = page
            val bitmap =
                pageFileStorage.getImage(page.pageId, PageFileStorage.PageFileType.ORIGINAL)

            bitmap?.let { image ->
                editPolygonImageView.setImageBitmap(image)
                editPolygonImageView.polygon = page.polygon
                val result = contourDetector.detect(image)
                editPolygonImageView.setLines(
                    result?.horizontalLines ?: emptyList(),
                    result?.verticalLines ?: emptyList()
                )
                editPolygonImageView.setMagnifier(magnifierView)
                magnifierView.setupMagnifier(editPolygonImageView)

            }
        }
        applyAppearanceConfig()
        croppingViewChannel.setMethodCallHandler(this)
    }

    private fun applyAppearanceConfig() {
        creationParams?.toCroppingViewConfig(gson)?.let { config ->
            config.edgeLineWidth?.let {
                editPolygonImageView.setEdgeWidth(convertDpToPx(context, it))
            }
            config.edgeColor?.let {
                editPolygonImageView.setEdgeColor(Color.parseColor(it))
            }
            config.anchorPointsColor?.let {
                editPolygonImageView.setAnchorPointsColor(Color.parseColor(it))
            }

            config.edgeColorOnLine?.let {
                editPolygonImageView.setEdgeColorOnLine(Color.parseColor(it))
            }

            config.borderInsets?.let { inset ->
                editPolygonImageView.setPadding(
                    inset.left?.let { value -> convertDpToPx(context, value).toInt() } ?: 0,
                    inset.top?.let { value -> convertDpToPx(context, value).toInt() } ?: 0,
                    inset.right?.let { value -> convertDpToPx(context, value).toInt() } ?: 0,
                    inset.bottom?.let { value -> convertDpToPx(context, value).toInt() } ?: 0,
                )
            }
        }
    }


    override fun onMethodCall(methodCall: MethodCall, result: MethodChannel.Result) {
        when (methodCall.method) {
            "rotateCW" -> {
                editPolygonImageView.rotateClockwise()
                result.success("success")
            }
            "rotateCCW" -> {
                editPolygonImageView.rotateCounterClockwise()
                result.success("success")
            }

            "reset" -> {
                editPolygonImageView.polygon = PolygonHelper.getFullPolygon()
                result.success("success")
            }

            "getEditedPolygon" -> {
                val rotation = editPolygonImageView.rotation
                val times = rotation / (-90);
                val polygon = editPolygonImageView.polygon
                result.success(getJsonForPolygonData(polygon, times))
            }
            "detect" -> {
                launch {
                    detectAndSetPolygon(result)
                }
            }
            "dispose" -> {
                dispose()
                result.success("success")
            }
            else -> result.notImplemented()
        }
    }

    private suspend fun CoroutineScope.detectAndSetPolygon(result: MethodChannel.Result) {
        page?.let {
            DetectPolygonOnPageUseCase(
                contourDetector,
                pageFileStorage
            ).detectPolygonOnPage(
                it
            ).onEach { data ->
                withContext(Dispatchers.Main) {
                    croppingViewChannel.invokeMethod("onShowHeavyProcessing", true)
                }
                val usePolygonFromPage = page?.polygon?.isNotEmpty() ?: false
                val detectedPolygon =
                    if (usePolygonFromPage) page?.polygon else data?.polygon
                editPolygonImageView.polygon =
                    detectedPolygon ?: PolygonHelper.getFullPolygon()
                editPolygonImageView.setLines(
                    data?.horizontal ?: emptyList(),
                    data?.vertical ?: emptyList()
                )
                result.success("success")
                withContext(Dispatchers.Main) {
                    croppingViewChannel.invokeMethod("onShowHeavyProcessing", false)
                }

            }.catch {
                result.success("error: ${it.localizedMessage}")
                withContext(Dispatchers.Main) {
                    croppingViewChannel.invokeMethod("onShowHeavyProcessing", true)
                }

            }.launchIn(this)
        }
    }

    private fun getJsonForPolygonData(
        polygon: List<PointF>,
        rotation: Float
    ): String {
        return gson.toJson(CroppingResult(polygon, rotation.toInt()))
    }


    override fun dispose() {
        job.cancel()
    }

    override fun onFlutterViewDetached() {
        super.onFlutterViewDetached()
    }

    override fun onFlutterViewAttached(flutterView: View) {
        super.onFlutterViewAttached(flutterView)
    }

}
